/*	WinFileCrypt.cpp
*	Justin Brady
*	Encrypts/Decrypts files using 3DES EDE with Cipher Block Chaining
*	Checksums with MD5
*	Made possible by crypto++ (5.0)
*/

#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <fstream.h>
#include <time.h>
#include <windows.h>
#include <winbase.h>

#include <cryptlib/cryptlib.h>
#include <cryptlib/des.h>
#include <cryptlib/misc.h>
#include <cryptlib/filters.h>
#include <cryptlib/modes.h>
#include <cryptlib/md5.h>

#include "WFCEncryptDecryptThread.h"

#include "WFCUtil.h"
#include "WFC_defines.h"


using namespace CryptoPP;


typedef unsigned char skey[KEYLEN];
typedef unsigned char hash[HASH_SIZE];	//128 bit hash


//function prototypes
int	__stdcall	WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd);
void	WFCInit();
LRESULT	CALLBACK	WFCWinProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void	HandleButtonActions(int btnId);
void	HandleFileDrop(WPARAM wParam);
bool	HandleConfirmKey();
void	ToggleEncryptDecrypt();
unsigned	long	WINAPI	DoEncryptDecrypt(void*);
void	DoQuit();
void DrawEncryptDecryptBitmap(HDC devContext, const POINT* pos);
void DrawLogo(HDC devContext);
unsigned long WINAPI WorkThread(LPVOID parms);
void DoEncryptDecryptWork();
void LogMessage(bool b, char* str);


//globals
HINSTANCE gInstanceHdl;
WNDCLASSEX outputWinClass;
HWND outputWin, dialogWin, ekeyEdit, infileEdit, bencryptBtn, goBtn, listBox;
HBITMAP encryptBitmap, decryptBitmap, logoBitmap;
HDC encryptBmpDC, decryptBmpDC, logoBmpDC;
RECT outputWin_rect;
HDC hOutputDevContext;
RECT ekeyEdit_rect;
LPCTSTR winClassName;
POINT curs;
POINT progbar_pos = {25, 170};
const POINT encryptdecryptBitmap_pos = {135, 140};
HBRUSH clearBrush;
HANDLE workThreadHdl;
DWORD workThreadId;
DWORD mainThreadId;

enum {
	outputWin_id = 0,
	ekeyEdit_id,
	infileEdit_id,
	bencryptBtn_id,
	goBtn_id,
	listBox_id,
};
		
bool done = false;
bool ekeyLocked = false;
int threadState = THREAD_WAITING;


bool	encrypt;
char*	argv[5];

//encrypt/decrypt thread
WFCEncryptDecryptThread *dEncryptDecryptThread;


int __stdcall WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
	MSG msg;

	gInstanceHdl = hInstance;

	char* inFileName = lpCmdLine;
	if(inFileName == NULL || strlen(inFileName) < 1)
		inFileName = "<Drag and drop files here>";

	char* nqInFileName = NoQuotes(inFileName);

	//is file encrypted?
	if(strstr(nqInFileName, CRYPT_FILEEXT) != NULL)
		encrypt = false;
	else
		encrypt = true;

	WFCInit();

	mainThreadId = GetCurrentThreadId();

	dEncryptDecryptThread = new WFCEncryptDecryptThread();
	dEncryptDecryptThread->SetUpdateProgressBar(&progbar_pos, outputWin);


	ListBoxAddString(listBox, "WinFileCrypt1.5 Started");

	SetWindowText(infileEdit, nqInFileName);

	do{
		//main window
		if(GetMessage(&msg, NULL, 0, 0))
		{
			if(threadState == THREAD_COMPLETE)
			{
				if(dEncryptDecryptThread->LastError() != 0)
				{
					char str[500];
					dEncryptDecryptThread->GetLastErrorMessage(str, 500);
					LogMessage(true, str); //WFCAssert(true, str, gInstanceHdl, outputWin);
					threadState = THREAD_WAITING;
				}
				else
				{
					if(encrypt)
						LogMessage(true, "Encrypt successful!");
					else
						LogMessage(true, "Decrypt successful!");

					ekeyLocked = true;	//because an encrypt/decrypt was successful, don't require confirmation
						
					threadState = THREAD_WAITING;
				}
			}

			if(msg.message == WM_KEYDOWN)
			{
				TranslateMessage(&msg);
				continue;
			}

			if(msg.message == WM_KEYUP)
			{
				TranslateMessage(&msg);
				continue;
			}

			if(msg.message == WM_QUIT)
				done = true;

			DispatchMessage(&msg);
		}
	}while(!done);

	//clean up
	DeleteDC(encryptBmpDC);
	DeleteDC(decryptBmpDC);
	DeleteDC(logoBmpDC);

	DeleteObject(encryptBitmap);
	DeleteObject(decryptBitmap);
	DeleteObject(logoBitmap);

	return 0;
}


void WFCInit()
{
	argv[1] = new char[WFC_BUFSZ1];
	argv[2] = new char[WFC_BUFSZ1];
	argv[3] = new char[WFC_BUFSZ1];

	outputWinClass.cbSize = sizeof(WNDCLASSEX);
	outputWinClass.style = CS_HREDRAW | CS_OWNDC;
	outputWinClass.lpfnWndProc = WFCWinProc;
	outputWinClass.cbClsExtra = 0;
	outputWinClass.cbWndExtra = DLGWINDOWEXTRA;
	outputWinClass.hInstance = gInstanceHdl;
	outputWinClass.hIcon = NULL;
	outputWinClass.hCursor = LoadCursor(NULL,IDC_ARROW);
	outputWinClass.hbrBackground = (HBRUSH) GetSysColorBrush(COLOR_3DFACE);
	outputWinClass.lpszMenuName = APPNAME;
	outputWinClass.lpszClassName = APPNAME;
	outputWinClass.hIconSm = NULL;

	RegisterClassEx(&outputWinClass);


	clearBrush = outputWinClass.hbrBackground;


	outputWin_rect.top = 100; outputWin_rect.left = 100;
	outputWin_rect.bottom = outputWin_rect.top + APPWIN_WINSIZEY; outputWin_rect.right = outputWin_rect.left + APPWIN_WINSIZEX;
	
	outputWin = CreateWindow(APPNAME, APPNAME, WS_OVERLAPPEDWINDOW | WS_VISIBLE,
		100, 100, APPWIN_WINSIZEX, APPWIN_WINSIZEY, NULL, NULL, gInstanceHdl, NULL);
	SetWindowFontDefault(outputWin);


	infileEdit = CreateWindow("edit", "",
		WS_VISIBLE | WS_BORDER | WS_CHILD | ES_LEFT | ES_AUTOHSCROLL,
		25, 50, 200, 20, outputWin, (HMENU) infileEdit_id, gInstanceHdl, NULL);
	SetWindowFontDefault(infileEdit);


	ekeyEdit = CreateWindow("edit", "",
		WS_VISIBLE | WS_BORDER | WS_CHILD | ES_LEFT | ES_PASSWORD,
		25, 100, 200, 20, outputWin, (HMENU) ekeyEdit_id, gInstanceHdl, NULL);
	SetWindowFontDefault(infileEdit);


	bencryptBtn = CreateWindow("button", encrypt? "Encrypt": "Decrypt",
		WS_VISIBLE | WS_BORDER | WS_CHILD,
		40, 135, 75, 30, outputWin, (HMENU) bencryptBtn_id, gInstanceHdl, NULL);
	SetWindowFontDefault(bencryptBtn);

	goBtn = CreateWindow("button", "Go!",
		WS_VISIBLE | WS_BORDER | WS_CHILD,
		90, 200, 75, 30, outputWin, (HMENU) goBtn_id, gInstanceHdl, NULL);
	SetWindowFontDefault(goBtn);

	listBox = CreateWindow("listbox", "Messages",
		WS_VISIBLE | WS_BORDER | WS_CHILD | WS_VSCROLL | ES_AUTOVSCROLL,
		25, 275, 200, 50, outputWin, (HMENU) listBox_id, gInstanceHdl, NULL);
	SetWindowFontDefault(listBox);

	SetBkColor(GetDC(outputWin), GetSysColor(COLOR_3DFACE));

	curs.x = 100;
	curs.y = 175;

	//load 'encrypt' bitmap
	encryptBitmap = LoadBitmap(gInstanceHdl, MAKEINTRESOURCE(WFC_RESID_ENCRYPTBITMAP));
	LogMessage(encryptBitmap == NULL, "Failed to load resource 103"); //WFCAssert(encryptBitmap == NULL, "Failed to load resource 103", gInstanceHdl, outputWin);

	encryptBmpDC = CreateCompatibleDC(hOutputDevContext);
	SelectObject(encryptBmpDC, encryptBitmap);

	//load 'decrypt' bitmap
	decryptBitmap = LoadBitmap(gInstanceHdl, MAKEINTRESOURCE(WFC_RESID_DECRYPTBITMAP));
	LogMessage(decryptBitmap == NULL, "Failed to load resource 104"); //WFCAssert(decryptBitmap == NULL, "Failed to load resource 104", gInstanceHdl, outputWin);

	decryptBmpDC = CreateCompatibleDC(hOutputDevContext);
	SelectObject(decryptBmpDC, decryptBitmap);

	//load logo
	logoBitmap = LoadBitmap(gInstanceHdl, MAKEINTRESOURCE(WFC_RESID_LOGOBITMAP));
	LogMessage(logoBitmap == NULL, "Failed to load resource 102"); //WFCAssert(logoBitmap == NULL, "Failed to load resource 102", gInstanceHdl, outputWin);

	logoBmpDC = CreateCompatibleDC(hOutputDevContext);
	SelectObject(logoBmpDC, logoBitmap);
}


//WinFileCrypt's WinProc
LRESULT CALLBACK WFCWinProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT paintStruct;
	HDC hrepaint;

	switch (uMsg)
	{
	case WM_CREATE:
		{
			DragAcceptFiles(hwnd, true);
		}break;

	case WM_PAINT:
		{
			BeginPaint(hwnd, &paintStruct);
			{
				hrepaint = paintStruct.hdc;

				//draw window here
				RECT fileNameRect, encryptKeyRect;
				
				SetRect(&fileNameRect, 25, 35, 100, 50);
				ExtTextOut(hrepaint, fileNameRect.left, fileNameRect.top, NULL, &fileNameRect, STR_FILENAME, strlen(STR_FILENAME), NULL);
				
				SetRect(&encryptKeyRect, 25, 85, 125, 100);
				ExtTextOut(hrepaint, encryptKeyRect.left, encryptKeyRect.top, NULL, &encryptKeyRect, STR_ENCRYPTKEY, strlen(STR_ENCRYPTKEY), NULL);
				
				dEncryptDecryptThread->UpdateProgress();

				//draw bitmap
				DrawEncryptDecryptBitmap(hrepaint, &encryptdecryptBitmap_pos);

				DrawLogo(hrepaint);
			}
			EndPaint(hwnd,&paintStruct);
		}break;

	case WM_DESTROY:
		{
			PostQuitMessage(0);
		}break;

	case WM_CLOSE:
		{
			DoQuit();
		}break;

	case WM_COMMAND:
		{			
			switch(HIWORD(wParam))
			{
			case BN_CLICKED:
				{
					HandleButtonActions(LOWORD(wParam));
				}break;

			case EN_SETFOCUS:
				{
					if((HWND) lParam == ekeyEdit)
					ekeyLocked = false;
				} break;

			case EN_KILLFOCUS:
				{
				} break;
				
			default:
				break;
			}
		}break;

	case WM_DROPFILES:
		{
			HandleFileDrop(wParam);
		}break;

	default:
		break;
	}

	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}


void HandleButtonActions(int btnId)
{
	char errorMsg[256];

	switch(btnId)
	{
	case bencryptBtn_id:
		ToggleEncryptDecrypt();
		break;

	case goBtn_id:
		{
			if(!ekeyLocked && encrypt) {
				bool b = HandleConfirmKey();
				if(!b)
					return;
			}

			GetTextFromControl(ekeyEdit, argv[1], WFC_BUFSZ1);

			GetTextFromControl(infileEdit, argv[2], WFC_BUFSZ1);

			strcpy(argv[3], argv[2]);

			if(encrypt)
				strcat(argv[3], CRYPT_FILEEXT);	//outfile is infile, with '.crypt' on the end
			else
			{
				char* ptr = strstr(argv[3], CRYPT_FILEEXT);	//strip off '.crypt' extension
				if(ptr != NULL)
					*ptr = '\0';
			}

			argv[0] = "UNUSED";

			//init
			LogMessage(true, "Initializing..");
			dEncryptDecryptThread->init(encrypt, argv);
			if(dEncryptDecryptThread->LastError() != 0) {
				dEncryptDecryptThread->GetLastErrorMessage(errorMsg, 256);
				LogMessage(true, errorMsg); //WFCAssert(true, errorMsg, gInstanceHdl, outputWin);
				return;
			}

			//begin
			if(encrypt)
				LogMessage(true, "Encrypt started...");
			else
				LogMessage(true, "Decrypt started...");
			
			DoEncryptDecryptWork();
		}
		break;
	
	default:
		break;
	}
}

void HandleFileDrop(WPARAM wParam)
{
	char filename[WFC_BUFSZ1];

	int n = DragQueryFile((HDROP) wParam, /*index in file list*/ 0, filename, WFC_BUFSZ1);

	SetWindowText(infileEdit, filename);

	char *ptr = filename + strlen(filename) - strlen(CRYPT_FILEEXT);

	if(strcmp(ptr, CRYPT_FILEEXT) == 0) {
		if(encrypt)
			ToggleEncryptDecrypt();
	}
	else
		if(!encrypt)
			ToggleEncryptDecrypt();

	LogMessage(true, "New file...");
}

bool HandleConfirmKey()
{
	bool b;

	if(ekeyLocked == false)
	{
		char *buf = new char[WFC_BUFSZ1];
		
		GetTextFromControl(ekeyEdit, buf, WFC_BUFSZ1);
		
		b = WFCConfirmKey(gInstanceHdl, outputWin, buf);
		if(b)
		{
			ekeyLocked = true;
			LogMessage(true, "Encrypt key confirmed");
		}
		else
			LogMessage(true, "Encrypt keys do not match!"); //WFCAssert(true, "Encrypt keys do not match!", gInstanceHdl, outputWin);
		
		delete buf;
	}

	return b;
}

void ToggleEncryptDecrypt()
{
	encrypt = !encrypt;
	
	if(encrypt)
		SetWindowText(bencryptBtn, "Encrypt");
	else
		SetWindowText(bencryptBtn, "Decrypt");

	DrawEncryptDecryptBitmap(GetDC(outputWin), &encryptdecryptBitmap_pos);
}


void DoQuit()
{
	PostQuitMessage(0);
	done = true;
}

void DrawEncryptDecryptBitmap(HDC devContext, const POINT* pos)
{
	if(encrypt) {
		BitBlt(devContext, pos->x, pos->y, WFC_BITMAP_ENCRYPTDECRYPT_SIZEX, WFC_BITMAP_ENCRYPTDECRYPT_SIZEY, encryptBmpDC, 0, 0, SRCCOPY);
	}

	else {
		BitBlt(devContext, pos->x, pos->y, WFC_BITMAP_ENCRYPTDECRYPT_SIZEX, WFC_BITMAP_ENCRYPTDECRYPT_SIZEY, decryptBmpDC, 0, 0, SRCCOPY);
	}
}

void DrawLogo(HDC devContext)//always draws at 0,0
{
	BitBlt(devContext, 0, 0, WFC_BITMAP_LOGO_SIZEX, WFC_BITMAP_LOGO_SIZEY, logoBmpDC, 0, 0, SRCCOPY);
}

unsigned long WINAPI WorkThread(LPVOID parms)
{
	//do encryptdecrypt
	dEncryptDecryptThread->DoEncryptDecrypt();

	threadState = THREAD_COMPLETE;

	return 0;
}

void DoEncryptDecryptWork()
{
	threadState = THREAD_RUNNING;
	workThreadHdl = CreateThread(NULL, 0, WorkThread, NULL, 0x0, &workThreadId);
}

void LogMessage(bool b, char* str)
{
	if(b)
	{
		ListBoxAddString(listBox, str);
		int n = SendMessage(listBox, LB_GETCOUNT, NULL, NULL);
		SendMessage(listBox, LB_SETCURSEL, n-1, NULL);
	}
}